@@ -15,6 +15,7 @@ import click
 
 import pynitrokey
 import pynitrokey.fido2.operations
+from pynitrokey.cli.pro import pro
 from pynitrokey.cli.fido2 import fido2
 from pynitrokey.cli.nethsm import nethsm
 from pynitrokey.cli.start import start
@@ -52,6 +53,7 @@ nitropy.add_command(fido2)
 nitropy.add_command(nethsm)
 nitropy.add_command(start)
 nitropy.add_command(storage)
+nitropy.add_command(pro)
 
 
 @click.command()
